"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class ScheduledRecording1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:ScheduledRecording:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'ScheduledRecording1'

    SERVICE_DEFAULT_VARIABLES = {
        "SortCapabilities": { "data_type": "string", "default": None, "allowed_list": None},
        "SortLevelCapability": { "data_type": "ui4", "default": None, "allowed_list": None},
        "StateUpdateID": { "data_type": "ui4", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {
        "LastChange": { "data_type": "string", "default": None, "allowed_list": None},
    }

    def action_BrowseRecordSchedules(self, Filter, StartingIndex, RequestedCount, SortCriteria, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the BrowseRecordSchedules action.

            :returns: "Result", "NumberReturned", "TotalMatches", "UpdateID"
        """
        arguments = {
            "Filter": Filter,
            "StartingIndex": StartingIndex,
            "RequestedCount": RequestedCount,
            "SortCriteria": SortCriteria,
        }

        out_params = self.call_action("BrowseRecordSchedules", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Result", "NumberReturned", "TotalMatches", "UpdateID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_BrowseRecordTasks(self, RecordScheduleID, Filter, StartingIndex, RequestedCount, SortCriteria, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the BrowseRecordTasks action.

            :returns: "Result", "NumberReturned", "TotalMatches", "UpdateID"
        """
        arguments = {
            "RecordScheduleID": RecordScheduleID,
            "Filter": Filter,
            "StartingIndex": StartingIndex,
            "RequestedCount": RequestedCount,
            "SortCriteria": SortCriteria,
        }

        out_params = self.call_action("BrowseRecordTasks", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Result", "NumberReturned", "TotalMatches", "UpdateID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_CreateRecordSchedule(self, Elements, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the CreateRecordSchedule action.

            :returns: "RecordScheduleID", "Result", "UpdateID"
        """
        arguments = {
            "Elements": Elements,
        }

        out_params = self.call_action("CreateRecordSchedule", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("RecordScheduleID", "Result", "UpdateID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_DeleteRecordSchedule(self, RecordScheduleID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the DeleteRecordSchedule action.
        """
        arguments = {
            "RecordScheduleID": RecordScheduleID,
        }

        self.call_action("DeleteRecordSchedule", arguments=arguments, aspects=aspects)

        return

    def action_DeleteRecordTask(self, RecordTaskID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the DeleteRecordTask action.
        """
        arguments = {
            "RecordTaskID": RecordTaskID,
        }

        self.call_action("DeleteRecordTask", arguments=arguments, aspects=aspects)

        return

    def action_DisableRecordSchedule(self, RecordScheduleID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the DisableRecordSchedule action.
        """
        arguments = {
            "RecordScheduleID": RecordScheduleID,
        }

        self.call_action("DisableRecordSchedule", arguments=arguments, aspects=aspects)

        return

    def action_DisableRecordTask(self, RecordTaskID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the DisableRecordTask action.
        """
        arguments = {
            "RecordTaskID": RecordTaskID,
        }

        self.call_action("DisableRecordTask", arguments=arguments, aspects=aspects)

        return

    def action_EnableRecordSchedule(self, RecordScheduleID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the EnableRecordSchedule action.
        """
        arguments = {
            "RecordScheduleID": RecordScheduleID,
        }

        self.call_action("EnableRecordSchedule", arguments=arguments, aspects=aspects)

        return

    def action_EnableRecordTask(self, RecordTaskID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the EnableRecordTask action.
        """
        arguments = {
            "RecordTaskID": RecordTaskID,
        }

        self.call_action("EnableRecordTask", arguments=arguments, aspects=aspects)

        return

    def action_GetAllowedValues(self, DataTypeID, Filter, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetAllowedValues action.

            :returns: "PropertyInfo"
        """
        arguments = {
            "DataTypeID": DataTypeID,
            "Filter": Filter,
        }

        out_params = self.call_action("GetAllowedValues", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("PropertyInfo",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetPropertyList(self, DataTypeID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetPropertyList action.

            :returns: "PropertyList"
        """
        arguments = {
            "DataTypeID": DataTypeID,
        }

        out_params = self.call_action("GetPropertyList", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("PropertyList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetRecordSchedule(self, RecordScheduleID, Filter, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetRecordSchedule action.

            :returns: "Result", "UpdateID"
        """
        arguments = {
            "RecordScheduleID": RecordScheduleID,
            "Filter": Filter,
        }

        out_params = self.call_action("GetRecordSchedule", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Result", "UpdateID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetRecordScheduleConflicts(self, RecordScheduleID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetRecordScheduleConflicts action.

            :returns: "RecordScheduleConflictIDList", "UpdateID"
        """
        arguments = {
            "RecordScheduleID": RecordScheduleID,
        }

        out_params = self.call_action("GetRecordScheduleConflicts", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("RecordScheduleConflictIDList", "UpdateID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetRecordTask(self, RecordTaskID, Filter, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetRecordTask action.

            :returns: "Result", "UpdateID"
        """
        arguments = {
            "RecordTaskID": RecordTaskID,
            "Filter": Filter,
        }

        out_params = self.call_action("GetRecordTask", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Result", "UpdateID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetRecordTaskConflicts(self, RecordTaskID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetRecordTaskConflicts action.

            :returns: "RecordTaskConflictIDList", "UpdateID"
        """
        arguments = {
            "RecordTaskID": RecordTaskID,
        }

        out_params = self.call_action("GetRecordTaskConflicts", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("RecordTaskConflictIDList", "UpdateID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetSortCapabilities(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetSortCapabilities action.

            :returns: "SortCaps", "SortLevelCap"
        """
        arguments = { }

        out_params = self.call_action("GetSortCapabilities", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("SortCaps", "SortLevelCap",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetStateUpdateID(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetStateUpdateID action.

            :returns: "Id"
        """
        arguments = { }

        out_params = self.call_action("GetStateUpdateID", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Id",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_ResetRecordTask(self, RecordTaskID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ResetRecordTask action.
        """
        arguments = {
            "RecordTaskID": RecordTaskID,
        }

        self.call_action("ResetRecordTask", arguments=arguments, aspects=aspects)

        return



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import os

from akit.exceptions import AKitSemanticError

class TestRootType:
    TESTPLUS = "testplus"

def lookup_test_root_type(test_root):

    test_root_module = os.path.join(test_root, os.path.join("__testroot__.py"))
    if not os.path.exists(test_root_module):
        errmsg = "The test root must have a module '__testroot__.py'. testroot={}".format(test_root_module)
        raise AKitSemanticError(errmsg) from None

    ROOT_TYPE = None

    trm_content = ""
    with open(test_root_module, 'r') as trmf:
        trm_content = trmf.read().strip()

    locals_vars = {}
    exec(trm_content, None, locals_vars)
    if "ROOT_TYPE" in locals_vars:
        ROOT_TYPE = locals_vars["ROOT_TYPE"]

    if ROOT_TYPE is None:
        errmsg = "The test root module must have a 'ROOT_TYPE' variable specifying (testplus, unittest).".format(test_root_module)
        raise AKitSemanticError(errmsg) from None

    if not ((ROOT_TYPE == TestRootType.TESTPLUS) or (ROOT_TYPE == TestRootType.UNITTEST)):
        errmsg = "Unknow test root type {}".format(ROOT_TYPE)
        raise AKitSemanticError(errmsg) from None

    return ROOT_TYPE



"""
.. module:: dnsvalidation
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing method used to validate DNS record information in accordance
               with RFC6763.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>

"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from akit.interop.dns.dnsconst import HAS_A_TO_Z
from akit.interop.dns.dnsconst import HAS_ONLY_A_TO_Z_NUM_HYPHEN
from akit.interop.dns.dnsconst import HAS_ASCII_CONTROL_CHARS
from akit.interop.dns.dnsconst import HAS_ONLY_A_TO_Z_NUM_HYPHEN_UNDERSCORE

from akit.interop.dns.dnserrors import DnsBadTypeInNameException

def validate_service_name(service_name: str, allow_underscores: bool = False):
    """
        Validates a service name for a DNS service.  The service name must meet the following
        requirements as per RFC6335

        1. Must begin with an underscore character '_'
        2. MUST be at least 1 character and no more than 15 characters long (not counting the mandatory underscore)
        3. MUST contain only US-ASCII [ANSI.X3.4-1986] letters 'A' - 'Z' and
           'a' - 'z', digits '0' - '9', and hyphens ('-', ASCII 0x2D or
           decimal 45)
        4. MAY not use consecutive hyphen characters '-'.
        5. MAY not use the hyphen character '-' as the first character after the underscore.
    """

    if service_name[0] != '_':
        raise DnsBadTypeInNameException("Service name (%s) must start with '_'" % service_name)

    # remove leading underscore
    service_name = service_name[1:]

    if len(service_name) > 15:
        raise DnsBadTypeInNameException("Service name (%s) must be <= 15 bytes" % service_name)

    if '--' in service_name:
        raise DnsBadTypeInNameException("Service name (%s) must not contain '--'" % service_name)

    if '-' in (service_name[0], service_name[-1]):
        raise DnsBadTypeInNameException("Service name (%s) may not start or end with '-'" % service_name)

    if not HAS_A_TO_Z.search(service_name):
        raise DnsBadTypeInNameException("Service name (%s) must contain at least one letter (eg: 'A-Z')" % service_name)

    allowed_characters_re = (
        HAS_ONLY_A_TO_Z_NUM_HYPHEN_UNDERSCORE if allow_underscores else HAS_ONLY_A_TO_Z_NUM_HYPHEN
    )

    if not allowed_characters_re.search(service_name):
        raise DnsBadTypeInNameException(
            "Service name (%s) must contain only these characters: "
            "A-Z, a-z, 0-9, hyphen ('-')%s" % (service_name, ", underscore ('_')" if allow_underscores else "")
        )

    return



"""
.. module:: eventedvariablesink
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the :class:`EventedVariableSink` class which is used to act
               as a container for the evented variables for the instance of a class representing
               the state of a remote object that propagates evented variable values.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>

"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2022, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Any, Dict, Optional, Tuple, Union

import threading
import weakref

from datetime import datetime
from akit.exceptions import AKitNotOverloadedError

from akit.interop.eventsinking.eventedvariable import EventedVariable

class EventedVariableSink:
    """
        The :class:`EventedVariableSink` provides a mechanism for managing 
    """

    SINK_VARIABLE_TYPE = EventedVariable


    def __init__(self, variable_description_table: dict, state_lock: Optional[threading.RLock]=None, sink_prefix: str=None, auto_subscribe: bool=False):
        super().__init__()

        self._event_state_lock = state_lock
        if self._event_state_lock is None:
            self._event_state_lock = threading.RLock()

        self._sink_prefix = sink_prefix
        self._variable_description_table = variable_description_table

        self._auto_subscribe = auto_subscribe
        self._subscription_id = None
        self._subscription_expiration = None

        self._evented_variables: Dict[str, EventedVariable] = {}
        self._initiator_moment_register: Dict[str, Union[Tuple[datetime, Any], Tuple[None, None]]] = {}

        self._create_event_variables_from_list()

        return

    @property
    def auto_subscribe(self):
        return self._auto_subscribe

    @property
    def subscriptionId(self) -> str:
        """
            Returns the subscription ID of the current subscription.
        """
        return self._subscription_id

    @property
    def subscriptionExpiration(self) -> str:
        return self._subscription_expiration

    def initiator_moment_lookup(self, event_name: str) -> Union[Tuple[datetime, Any], Tuple[None, None]]:
        """
            Lookup the time of initiation for the event specified and any associated context information.

            :param event_name: The event identifier for the initiation moment being looked up.

            :returns: A tuple containing the last moment of initiation and context for the event specified.
        """

        rtnval = None, None

        for _ in self.yield_state_lock():
            if event_name in self._initiator_moment_register:
                rtnval = self._initiator_moment_register[event_name]

        return rtnval

    def initiator_moment_register(self, event_name: str, moment: datetime, context: Optional[Any]=None) -> None:
        """
            Registers the initiator time and context for a given initiator topic.

            :param event_name: The event identifier for the initiation moment being recorded.
            :param datetime: The datetime of the moment of initiation for a topic.
            :param context: Contextual information that provides more details around then initiation moment.
        """
        
        self._event_state_lock.acquire()
        try:
            self._initiator_moment_register[event_name] = (moment, context)
        finally:
            self._event_state_lock.release()

        return
    
    def lookup_event_variable(self, event_name: str) -> Union[EventedVariable, None]:
        """
            Looks up the specified event variable.

            :param event_name: The event name to find the :class:`EventedVariable` for.
        """
        varobj = None

        varkey = event_name
        if self._sink_prefix is not None:
            varkey = "{}/{}".format(self._sink_prefix, event_name)

        for _ in self.yield_state_lock():
            if varkey in self._evented_variables:
                varobj = self._evented_variables[varkey]

        return varobj

    def update_event_variable(self, event_name: str, event_value: Any, sink_locked: bool=False):

        varkey = event_name
        if self._sink_prefix is not None:
            varkey = "{}/{}".format(self._sink_prefix, event_name)

        if sink_locked:
            varobj = self._evented_variables[varkey]
            varobj.sync_update(event_value, expires=self._subscription_expiration, sink_locked=True)

        else:
            for _ in self.yield_state_lock():
                varobj = self._evented_variables[varkey]
                varobj.sync_update(event_value, expires=self._subscription_expiration, sink_locked=True)

        return

    def invalidate_subscription(self, scope: Optional[Any]=None):
        """
            Called in order to invalidate the subscription(s) specified by scope.

            :param scope: The scope of the subscriptions to renew.  If not specified then all
                          subscriptions should be renewed.
        """
        errmsg = "The `invalidate_subscription` method was not overloaded for type={}".format(type(self).__name__)
        raise AKitNotOverloadedError(errmsg)

    def renew_subscription(self, scope: Optional[Any]=None):
        """
            Called in order to renew the subscription(s) specified by scope. 

            :param scope: The scope of the subscriptions to renew.  If not specified then all
                          subscriptions should be renewed.
        """
        errmsg = "The `renew_subscription` method was not overloaded for type={}".format(type(self).__name__)
        raise AKitNotOverloadedError(errmsg)
    
    def trigger_auto_subscribe_from_variable(self, varkey: str):
        """
            Called in order to renew the subscription to the 

            :param varkey: The key for the variable that is triggering the auto-subscription process.
        """
        errmsg = "The `trigger_auto_subscribe_from_variable` method was not overloaded for type={}".format(type(self).__name__)
        raise AKitNotOverloadedError(errmsg)

    def yield_state_lock(self) -> threading.RLock:
        """
            Yields the state lock in a way that it can be automatically released at the end of an
            iteration scope.
        """
        self._event_state_lock.acquire()
        try:
            yield
        finally:
            self._event_state_lock.release()

    def _create_event_variable(self, event_name: str, **event_desc):
        """
            Creates an event variable and stores a reference to it in the variables table.

            :param event_name: The name of the event variable to create.
            :param data_type: The type of the event variable to create.
            :param default: The default value to set the new event variable to.
        """
        varkey = event_name
        if self._sink_prefix is not None:
            varkey = "{}/{}".format(self._sink_prefix, event_name)

        sink_ref = weakref.ref(self)
        event_var = self.SINK_VARIABLE_TYPE(varkey, event_name, sink_ref, **event_desc)
        self._evented_variables[varkey] = event_var

        return

    def _create_event_variables_from_list(self):
        """
            Called by the constructor to create the event variables listed from the variable descriptions
            detailed in the EVENTED_VARIABLE_DESCRIPTIONS table of the :class:`EventedVariableSink` derived type.
            We pre-create the event variables for each instance of an :class:`EventedVariableSink` type as the
            sink objects are designed to be used in a one to one, instance to remote object way.
        """

        for event_name in self._variable_description_table:
            event_desc = self._variable_description_table[event_name]
            self._create_event_variable(event_name, **event_desc)

        return




"""
.. module:: taskbase
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module which contains the :class:`TaskBase` object which is used as the base.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

class TaskBase:

    required_scopes = None

    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        self._scope = None
        return

    @classmethod
    def do(cls, scope, *args, **kwargs):
        return

    @classmethod
    def begin(cls, scope, *args, **kwargs):
        return

    def enter(self):
        self._scope.enter()
        return

    def exit(self):
        self._scope.exit()
        return



"""
.. module:: testnode
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module that contains the :class:`TestNode` class which is utilized as the collection point
               which associates a set of tests with their descendant execution scopes.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Dict, List, Optional

import sys

from akit.exceptions import AKitSemanticError

from akit.testing.testplus.registration.resourceregistry import resource_registry
from akit.testing.testplus.testref import TestRef

from akit.xlogging.foundations import getAutomatonKitLogger

logger = getAutomatonKitLogger()

class TestGroup:
    """
              -------------
              |  Group A  |
        ---------------------------
        |  Group AA  |  Scope AB  |
        -------------------------------
        |         Scope AAA/ABA       |
        -------------------------------
    """

    def __init__(self, name, package=None):
        self._name = name
        self._package = package
        self._children = {}
        self._finalized = False
        self._metadata = None
        return

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_inst, ex_tb):
        return False

    @property
    def children(self):
        return self._children

    @property
    def finalized(self):
        return self._finalized

    @property
    def name(self):
        return self._name

    @property
    def package(self):
        return self._package

    @property
    def scope_name(self):
        sname = self._name
        if self._package is not None and len(self._package) > 0:
            sname = "{}.{}".format(self.package, sname)
        return sname

    def add_descendent(self, test_ref:TestRef):

        if self._package is not None:
            err_msg = "The 'add_descendent' API can only be called on the root package."
            raise AKitSemanticError(err_msg) from None

        testname = test_ref.test_name
        module_name, _ = testname.split("#")
        to_walk_list = module_name.split(".")
        path_stack = []

        self._add_descendent(test_ref, to_walk_list, path_stack)

        return

    def finalize(self):
        self._finalized = True
        return

    def get_resource_scope(self):
        scope_name = self.scope_name
        rscope = resource_registry.lookup_resource_scope(scope_name)
        return rscope

    def resolve_metadata(self, parent_metadata: Optional[Dict[str, str]]=None):

        reference_metadata = self._reference_metadata()

        if parent_metadata is not None:
            if reference_metadata is not None:
                self._metadata = {}
                self._metadata.update(parent_metadata)
                self._metadata.update(reference_metadata)
            else:
                self._metadata = parent_metadata
        else:
            self._metadata = reference_metadata

        for child in self._children.values():
            child.resolve_metadata(self._metadata)

        return

    def _add_descendent(self, test_ref:TestRef, to_walk_list: List[str], path_stack: List[str],):
        
        if len(to_walk_list) == 0:
            tbname = test_ref.test_base_name
            self._children[tbname] = test_ref
        else:
            child_leaf = to_walk_list[0]

            desc_to_walk_list = []
            if len(to_walk_list) > 1:
                desc_to_walk_list = to_walk_list[1:]

            child_leaf_group = None
            if child_leaf in self._children:
                child_leaf_group = self._children[child_leaf]
            else:
                tgname = child_leaf
                tgpkg = ".".join(path_stack) 
                child_leaf_group = TestGroup(tgname, tgpkg)
                self._children[child_leaf] = child_leaf_group

            path_stack.append(child_leaf)
            try:
                child_leaf_group._add_descendent(test_ref, desc_to_walk_list, path_stack)
            finally:
                path_stack.pop()

        return

    def _reference_metadata(self):
        """
            Looks up the metadata if any on the module associated with this group.
        """
        
        modname = self._name
        if self._package is not None and self._package != "":
            modname = "{}.{}".format(self._package, self._name)

        refmd = None
        if modname in sys.modules:
            mod = sys.modules[modname]
            if hasattr(mod, "_metadata_"):
                refmd = mod._metadata_

        return refmd

    def __contains__(self, key):
        has_item = key in self._children
        return has_item

    def __getitem__(self, key):
        item = self._children[key]
        return item

    def __setitem__(self, key, value):
        self._children[key] = value
        return



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import List, NamedTuple, Union

from akit.xregex.regexreader import RegExReader, RegExPattern

class ProcessItem(NamedTuple):
    pid: int
    tty: str
    time: str
    cmd: str

class CmdReaderPs(RegExReader):
    """
        Reader for processing the output of the 'ps' command.
    """

    EXPECTED_LINES = [
        RegExPattern(
            r"^[\s]*(?P<hpid>[A-Za-z]+)[\s]+(?P<htty>[A-Za-z]+)[\s]+(?P<htime>[A-Za-z]+)[\s]+(?P<hcmd>[A-Za-z]+)",
            consume=True),
        RegExPattern(
            r"^[\s]*(?P<pid>[\S]+)[\s]+(?P<tty>[\S]+)[\s]+(?P<time>[\S]+)[\s]+(?P<cmd>[\S]+)",
            destination="process_listing", match_type=ProcessItem, repeats=True)
    ]

    def __init__(self, content:Union[List, str], strict=False):
        self._process_listing = []

        super().__init__(content, strict=strict)
        return

    @property
    def process_listing(self) -> List[ProcessItem]:
        return self._process_listing


if __name__ == "__main__":

    cmd_content = """
    PID TTY          TIME CMD
  82785 pts/3    00:00:00 bash
  88819 pts/3    00:00:00 ps
"""

    psrdr = CmdReaderPs(cmd_content)

    for pitem in psrdr.process_listing:
        print(pitem)



"""
.. module:: xenum
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module which contains functions for formatting text.

.. note:: The modules that are named `xsomething` like this module are prefixed with an `x` character to
          indicate they extend the functionality of a base python module and the `x` is pre-pended to
          prevent module name collisions with python modules.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from enum import Enum

class DocStringEnum(str, Enum):
    """
        Enumeration that assigns a string and an optional documentation string
        to a Label.
    """
    def __new__(cls, value, doc=None):
        obj = str.__new__(cls, value)
        obj._value_ = value
        if doc is not None:
            obj.__doc__ = doc
        return obj

class DocIntEnum(int, Enum):
    """
        Enumeration that assigns a string and an optional documentation string
        to a Label.
    """
    def __new__(cls, value, doc=None):
        obj = int.__new__(cls, value)
        obj._value_ = value
        if doc is not None:
            obj.__doc__ = doc
        return obj




"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class Scan1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:Scan:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'Scan1'

    SERVICE_DEFAULT_VARIABLES = {
        "AppendSideNumber": { "data_type": "string", "default": "0", "allowed_list": "['device-setting', '0']"},
        "BaseName": { "data_type": "string", "default": "pull-relative", "allowed_list": None},
        "BitDepth": { "data_type": "string", "default": "8", "allowed_list": "['device-setting', '8']"},
        "ColorSpace": { "data_type": "string", "default": "sRGB", "allowed_list": "['device-setting', 'sRGB']"},
        "ColorType": { "data_type": "string", "default": "Color", "allowed_list": "['device-setting', 'Color']"},
        "CompressionFactor": { "data_type": "i4", "default": "100", "allowed_list": None},
        "Destination": { "data_type": "string", "default": None, "allowed_list": None},
        "DeviceID": { "data_type": "string", "default": None, "allowed_list": None},
        "ErrorTimeout": { "data_type": "i4", "default": None, "allowed_list": None},
        "HeightLimit": { "data_type": "i4", "default": None, "allowed_list": None},
        "ImageFormat": { "data_type": "string", "default": "image/jpeg", "allowed_list": "['device-setting', 'image/jpeg']"},
        "ImageType": { "data_type": "string", "default": "Mixed", "allowed_list": "['device-setting', 'Mixed']"},
        "JobID": { "data_type": "ui4", "default": None, "allowed_list": None},
        "JobName": { "data_type": "string", "default": None, "allowed_list": None},
        "RegistrationID": { "data_type": "ui4", "default": None, "allowed_list": None},
        "Resolution": { "data_type": "string", "default": None, "allowed_list": None},
        "SideCount": { "data_type": "i4", "default": "0", "allowed_list": None},
        "StateReason": { "data_type": "string", "default": None, "allowed_list": None},
        "Timeout": { "data_type": "i4", "default": None, "allowed_list": None},
        "UseFeeder": { "data_type": "string", "default": "0", "allowed_list": "['device-setting', '0']"},
        "WidthLimit": { "data_type": "i4", "default": None, "allowed_list": None},
        "XValueLimit": { "data_type": "i4", "default": None, "allowed_list": None},
        "YValueLimit": { "data_type": "i4", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {
        "DestinationID": { "data_type": "ui4", "default": None, "allowed_list": None},
        "FailureCode": { "data_type": "string", "default": "No Error", "allowed_list": "['No Error', 'Jammed', 'Timeout Reached', 'ErredTimeout Reached', 'Destination Not Reachable']"},
        "ScanLength": { "data_type": "i4", "default": "0", "allowed_list": None},
        "SideNumber": { "data_type": "i4", "default": "1", "allowed_list": None},
        "State": { "data_type": "string", "default": "Idle", "allowed_list": "['Idle', 'Reserved', 'NotReady', 'Pending', 'Scanning', 'Finishing', 'Erred']"},
    }

    def action_GetConfiguration(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetConfiguration action.

            :returns: "JobNameOut", "ResolutionOut", "ImageXOffsetOut", "ImageYOffsetOut", "ImageWidthOut", "ImageHeightOut", "ImageFormatOut", "CompressionFactorOut", "ImageTypeOut", "ColorTypeOut", "BitDepthOut", "ColorSpaceOut", "BaseNameOut", "AppendSideNumberOut", "TimeoutOut"
        """
        arguments = { }

        out_params = self.call_action("GetConfiguration", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("JobNameOut", "ResolutionOut", "ImageXOffsetOut", "ImageYOffsetOut", "ImageWidthOut", "ImageHeightOut", "ImageFormatOut", "CompressionFactorOut", "ImageTypeOut", "ColorTypeOut", "BitDepthOut", "ColorSpaceOut", "BaseNameOut", "AppendSideNumberOut", "TimeoutOut",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetSideInformation(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetSideInformation action.

            :returns: "SideNumberOut", "SideCountOut", "ScanLengthOut"
        """
        arguments = { }

        out_params = self.call_action("GetSideInformation", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("SideNumberOut", "SideCountOut", "ScanLengthOut",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetState(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetState action.

            :returns: "StateOut", "StateReasonOut", "FailureCodeOut"
        """
        arguments = { }

        out_params = self.call_action("GetState", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("StateOut", "StateReasonOut", "FailureCodeOut",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_StartScan(self, RegistrationIDIn, UseFeederIn, SideCountIn, JobNameIn, ResolutionIn, ImageXOffsetIn, ImageYOffsetIn, ImageWidthIn, ImageHeightIn, ImageFormatIn, CompressionFactorIn, ImageTypeIn, ColorTypeIn, BitDepthIn, ColorSpaceIn, BaseNameIn, AppendSideNumberIn, TimeoutIn, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the StartScan action.

            :returns: "ActualTimeoutOut", "JobIDOut", "ActualWidthOut", "ActualHeightOut"
        """
        arguments = {
            "RegistrationIDIn": RegistrationIDIn,
            "UseFeederIn": UseFeederIn,
            "SideCountIn": SideCountIn,
            "JobNameIn": JobNameIn,
            "ResolutionIn": ResolutionIn,
            "ImageXOffsetIn": ImageXOffsetIn,
            "ImageYOffsetIn": ImageYOffsetIn,
            "ImageWidthIn": ImageWidthIn,
            "ImageHeightIn": ImageHeightIn,
            "ImageFormatIn": ImageFormatIn,
            "CompressionFactorIn": CompressionFactorIn,
            "ImageTypeIn": ImageTypeIn,
            "ColorTypeIn": ColorTypeIn,
            "BitDepthIn": BitDepthIn,
            "ColorSpaceIn": ColorSpaceIn,
            "BaseNameIn": BaseNameIn,
            "AppendSideNumberIn": AppendSideNumberIn,
            "TimeoutIn": TimeoutIn,
        }

        out_params = self.call_action("StartScan", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("ActualTimeoutOut", "JobIDOut", "ActualWidthOut", "ActualHeightOut",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args



"""
.. module:: extensible
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing classes that enable the discovery of extensions

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import List
from types import ModuleType

import inspect
import os

from akit.compat import import_file, import_by_name
from akit.paths import get_directory_for_code_container

class LoadableExtension: # pylint: disable=too-few-public-methods
    """
        Marks a class as an extension for collection purposes so we can distinguish
        extension classes from base classes
    """

def collect_extensions_under_code_container(container: ModuleType, ext_base_type: type) -> List[type]:
    """
        Scans the code `container` provide and all descendant containers for classes
        that inherit from the type passed as `ext_base_type`

        :param container: A python package or module to scan for extension types.
        :param ext_base_type: A python class type that serves as a base class to identify other
                              types that are a type of extension.

        :returns: A list of types found that inherit from `ext_base_type`
    """
    ext_collection = []

    # This is declare here so it can be used as a closure
    nxtmod = None

    def is_extension_class(obj):
        result = False

        if inspect.isclass(obj):
            obj_container = obj.__module__
            if obj_container == nxtmod.__name__ and LoadableExtension in obj.__bases__:
                result = issubclass(obj, ext_base_type) and obj is not ext_base_type
        return result

    container_name = container.__name__
    container_dir = get_directory_for_code_container(container)
    container_parts = container_name.split(".")
    container_root = os.sep.join(container_dir.split(os.sep)[:-len(container_parts)])
    rootlen = len(container_root)

    for dirpath, _, filenames in os.walk(container_dir):
        leafdir = dirpath[rootlen:].lstrip(os.sep)
        leafcontainer = leafdir.replace(os.sep, ".")
        for nxtfile in filenames:
            nfbase, nfext = os.path.splitext(nxtfile)
            if nfext != ".py":
                continue

            if nfbase == "__init__":
                nxtmodname = "%s" % (leafcontainer)
            else:
                nxtmodname = "%s.%s" % (leafcontainer, nfbase)

            nxtmod = import_by_name(nxtmodname)
            if nxtmod is None:
                continue

            ext_collection.extend(inspect.getmembers(nxtmod, predicate=is_extension_class))

    return ext_collection

def collect_extensions_under_folder(extension_folder: str, ext_base_type: type, module_base: str="akit.ext.generated") -> List[type]:
    """
        Scans the code `container` provide and all descendant containers for classes
        that inherit from the type passed as `ext_base_type`

        :param container: A python package or module to scan for extension types.
        :param ext_base_type: A python class type that serves as a base class to identify other
                              types that are a type of extension.

        :returns: A list of types found that inherit from `ext_base_type`
    """
    ext_collection = []

    # This is declare here so it can be used as a closure
    nxtmod = None

    def is_extension_class(obj):
        result = False

        if inspect.isclass(obj):
            obj_container = obj.__module__
            if obj_container == nxtmod.__name__ and LoadableExtension in obj.__bases__:
                result = issubclass(obj, ext_base_type) and obj is not ext_base_type
        return result

    rootlen = len(extension_folder)

    for dirpath, _, filenames in os.walk(extension_folder):
        leafdir = dirpath[rootlen:].lstrip(os.sep)
        leafcontainer = leafdir.replace(os.sep, ".")
        for nxtfile in filenames:
            filefull = os.path.join(dirpath, nxtfile)
            nfbase, nfext = os.path.splitext(nxtfile)
            if nfext != ".py":
                continue

            if nfbase == "__init__":
                continue
            
            nxtmodname = "%s.%s.%s" % (module_base, leafcontainer, nfbase)
            nxtmod = import_file(nxtmodname, filefull, by_file_only=True)
            if nxtmod is None:
                continue

            ext_collection.extend(inspect.getmembers(nxtmod, predicate=is_extension_class))

    return ext_collection

def generate_extension_key(*parts: str) -> str:
    """
        Generates a unique key that identifies an extension type based on where
        it was found in a hiearchy of code containers.

        :params parts: List of names of the path to the extension type

        :returns: A unique path based identifier for a type.
    """
    extkey = "/".join(parts)
    return extkey




__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from string import ascii_lowercase, ascii_uppercase, digits

# We might want to write any generated passwords into a log or config file
# so don't use all the printable characters.  The security offered by these
# charaters is sufficient for automation purposes
WPA_USEABLE_CHARACTERS =  ascii_lowercase + ascii_uppercase + digits

WPA_MIN_CHAR_COUNT = 8
WPA_MAX_CHAR_COUNT = 63

import random

def create_random_wpa_password():
    """
    """
    char_count = random.choice(range(WPA_MIN_CHAR_COUNT, WPA_MAX_CHAR_COUNT))
    key = random.sample(WPA_USEABLE_CHARACTERS, char_count)
    return key


"""
.. module:: xformatting
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module which contains functions for formatting text.

.. note:: The modules that are named `xsomething` like this module are prefixed with an `x` character to
          indicate they extend the functionality of a base python module and the `x` is pre-pended to
          prevent module name collisions with python modules.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import List, Optional

import os

from io import StringIO

class CommandOutputFormat:
    DISPLAY = 0
    JSON = 1
    YAML = 2

def format_command_result(msg: str, status: int, stdout: str, stderr: str, exp_status: Optional[List[int]]=None) -> str:
    """
        Takes a message and command results and formats a message for output to the logs.

        :param msg: The message to output to the logs
        :param status: The return status code associated with the command.
        :param stdout: The std out text from the command.
        :param stderr: The std error text from the command.

        :returns: The formatted message output.
    """

    fmt_msg_lines = [msg, "STATUS: {}".format(status)]
    
    if exp_status is not None:
        fmt_msg_lines.append("EXPECTED: {}".format(repr(exp_status)))

    if stdout is not None and len(stdout.strip()) > 0:
        fmt_msg_lines.append("STDOUT:")
        fmt_msg_lines.append(indent_lines(stdout, 1))

    if stderr is not None and len(stderr.strip()) > 0:
        fmt_msg_lines.append("STDERR:")
        fmt_msg_lines.append(indent_lines(stdout, 1))

    fmt_msg = os.linesep.join(fmt_msg_lines)

    return fmt_msg

def indent_lines(msg: str, level: int, indent: int=4) -> str:
    """
        Takes a string and splits it into multiple lines, then indents each line
        to the specified level using 'indent' spaces for each level.

        :param msg: The text content to split into lines and then indent.
        :param level: The integer level number to indent to.
        :param indent: The number of spaces to indent for each level.

        :returns: The indenting content
    """
    # Split msg into lines keeping the line endings
    msglines = msg.splitlines(True)

    pfx = " " * (level * indent)

    indented = StringIO()
    for nxtline in msglines:
        indented.write(pfx)
        indented.write(nxtline)

    return indented.getvalue()

def indent_line(lcontent: str, level: int, indent: int, pre_strip_leading: bool=True) -> str:
    """
        Takes a string and indents it to the specified level using 'indent' spaces
        for each level.

        :param lcontent: The text line to indent.
        :param level: The integer level number to indent to.
        :param indent: The number of spaces to indent for each level.
        :param pre_strip_leading: Strip any leading whitesspace before indenting the line.

        :returns: The indented line
    """
    pfx = " " * (level * indent)

    indented = None
    if pre_strip_leading:
        indented = "{}{}".format(pfx, lcontent.lstrip())
    else:
        indented = "{}{}".format(pfx, lcontent)

    return indented

def split_and_indent_lines(msg: str, level: int, indent: int=4, pre_strip_leading: bool=True) -> List[str]:
    """
        Takes a string and splits it into multiple lines, then indents each line
        to the specified level using 'indent' spaces for each level.

        :param msg: The text content to split into lines and then indent.
        :param level: The integer level number to indent to.
        :param indent: The number of spaces to indent for each level.
        :param pre_strip_leading: Strip any leading whitesspace before indenting the lines.

        :returns: The indenting lines
    """

    # Split msg into lines keeping the line endings
    msglines = msg.splitlines(False)

    pfx = " " * (level * indent)

    indented = None
    if pre_strip_leading:
        indented = [pfx + nxtline.lstrip() for nxtline in msglines]
    else:
        indented = [pfx + nxtline for nxtline in msglines]

    return indented



"""
.. module:: wirelessapcoordinator
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Contains the WirelessAPCoordinator which is used for managing wireless access points.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"


from typing import TYPE_CHECKING

from akit.exceptions import AKitConfigurationError
from akit.interop.coordinators.coordinatorbase import CoordinatorBase
from akit.interop.credentials.credentialmanager import CredentialManager

from akit.interop.agents.wirelessapagent import WirelessApAgent

if TYPE_CHECKING:
    from akit.interop.landscaping.landscape import Landscape

class WirelessAPCoordinator(CoordinatorBase):

    def __init__(self, lscape: "Landscape", *args, **kwargs):
        super(WirelessAPCoordinator, self).__init__(lscape, *args, **kwargs)
        return

    def _initialize(self, *_args, **_kwargs):
        """
            Called by the CoordinatorBase constructor to perform the one time initialization of the coordinator Singleton
            of a given type.
        """
        self._wireless_ap_config = {}
        for apcfg in self._coord_config:
            cfgname = apcfg["name"]
            self._wireless_ap_config[cfgname] = apcfg

        self._wireless_agents = {}
        return
    
    def lookup_agent(self, apname: str) -> WirelessApAgent:
        """
            Looks up a serial agent by serial mapping.
        """
        wireless_agent = None

        lscape = self.landscape

        if apname in self.self._wireless_ap_config:
            wireless_mapping = self._wireless_ap_config[apname]

            if apname not in self._wireless_agents:
                host = wireless_mapping["host"]
                credential_name = wireless_mapping["credential"]

                credential = lscape.lookup_credential(credential_name)

                if credential is not None:
                    wireless_agent = WirelessApAgent(host, credential)

                    self._wireless_agents[apname] = wireless_agent
                else:
                    errmsg = "Failure to find credential '{}' specified for wireless apname={}".format(
                        credential_name, apname
                    )
                    raise AKitConfigurationError(errmsg)
            else:
                wireless_agent = self._wireless_agents[apname]
        else:
            errmsg = "Failure to lookup wireless configuration for apname={}.".format(apname)
            raise AKitConfigurationError(errmsg) from None

        return wireless_agent



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class WANDSLLinkConfig1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:WANDSLLinkConfig:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'WANDSLLinkConfig1'

    SERVICE_DEFAULT_VARIABLES = {
        "ATMEncapsulation": { "data_type": "string", "default": None, "allowed_list": None},
        "AutoConfig": { "data_type": "boolean", "default": None, "allowed_list": None},
        "DestinationAddress": { "data_type": "string", "default": None, "allowed_list": None},
        "FCSPreserved": { "data_type": "boolean", "default": None, "allowed_list": None},
        "LinkStatus": { "data_type": "string", "default": None, "allowed_list": "['Up', 'Down']"},
        "LinkType": { "data_type": "string", "default": None, "allowed_list": None},
        "ModulationType": { "data_type": "string", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {}

    def action_GetATMEncapsulation(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetATMEncapsulation action.

            :returns: "NewATMEncapsulation"
        """
        arguments = { }

        out_params = self.call_action("GetATMEncapsulation", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewATMEncapsulation",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetAutoConfig(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetAutoConfig action.

            :returns: "NewAutoConfig"
        """
        arguments = { }

        out_params = self.call_action("GetAutoConfig", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewAutoConfig",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetDSLLinkInfo(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetDSLLinkInfo action.

            :returns: "NewLinkType", "NewLinkStatus"
        """
        arguments = { }

        out_params = self.call_action("GetDSLLinkInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewLinkType", "NewLinkStatus",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetDestinationAddress(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetDestinationAddress action.

            :returns: "NewDestinationAddress"
        """
        arguments = { }

        out_params = self.call_action("GetDestinationAddress", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewDestinationAddress",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetFCSPreserved(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetFCSPreserved action.

            :returns: "NewFCSPreserved"
        """
        arguments = { }

        out_params = self.call_action("GetFCSPreserved", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewFCSPreserved",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetModulationType(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetModulationType action.

            :returns: "NewModulationType"
        """
        arguments = { }

        out_params = self.call_action("GetModulationType", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewModulationType",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetATMEncapsulation(self, NewATMEncapsulation, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetATMEncapsulation action.
        """
        arguments = {
            "NewATMEncapsulation": NewATMEncapsulation,
        }

        self.call_action("SetATMEncapsulation", arguments=arguments, aspects=aspects)

        return

    def action_SetDSLLinkType(self, NewLinkType, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetDSLLinkType action.
        """
        arguments = {
            "NewLinkType": NewLinkType,
        }

        self.call_action("SetDSLLinkType", arguments=arguments, aspects=aspects)

        return

    def action_SetDestinationAddress(self, NewDestinationAddress, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetDestinationAddress action.
        """
        arguments = {
            "NewDestinationAddress": NewDestinationAddress,
        }

        self.call_action("SetDestinationAddress", arguments=arguments, aspects=aspects)

        return

    def action_SetFCSPreserved(self, NewFCSPreserved, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetFCSPreserved action.
        """
        arguments = {
            "NewFCSPreserved": NewFCSPreserved,
        }

        self.call_action("SetFCSPreserved", arguments=arguments, aspects=aspects)

        return



"""
.. module:: upnpserviceproxy
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the :class:`UpnpServiceProxy` class which is the base class
               all services proxied.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>

"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Optional, Union, TYPE_CHECKING

import os
import threading
import time
import traceback
import weakref

from datetime import datetime, timedelta

import requests

from akit.aspects import AspectsUPnP, ActionPattern, DEFAULT_UPNP_ASPECTS

from akit.environment.context import Context
from akit.environment.contextpaths import ContextPaths

from akit.xlogging.foundations import getAutomatonKitLogger
from akit.xlogging.scopemonitoring import MonitoredScope
from akit.xtraceback import format_exc_lines

from akit.exceptions import AKitRuntimeError, AKitTimeoutError

from akit.interop.eventsinking.eventedvariablesink import EventedVariableSink

from akit.interop.upnp.soap import SoapProcessor, SOAP_TIMEOUT
from akit.interop.upnp.upnpconstants import DEFAULT_UPNP_CALL_ASPECTS
from akit.interop.upnp.upnperrors import UpnpError
from akit.interop.upnp.services.upnpdefaultvar import UpnpDefaultVar

from akit.interop.eventsinking.eventedvariable import EventedVariable

from akit.interop.upnp.xml.upnpdevice1 import UpnpDevice1Service

from akit.timeouts import TimeoutContext

logger = getAutomatonKitLogger()

if TYPE_CHECKING:
    from akit.interop.upnp.devices.upnpdevice import UpnpDevice

class UpnpServiceProxy(EventedVariableSink):
    """
        The :class:`UpnpServiceProxy` object provides that base data and functional methods for providing
        inter-operability with a devices Service.  It provides methods for making calls on actions and
        access to methods for making subscriptions.
    """

    SERVICE_ID = None
    SERVICE_TYPE = None

    SERVICE_DEFAULT_VARIABLES = {}

    SERVICE_EVENT_VARIABLES = {}

    def __init__(self):

        self._service_lock = threading.RLock()

        self._device_ref = None
        self._soap_processor = SoapProcessor()

        self._host = None
        self._baseURL = None

        self._controlURL = None
        self._eventSubURL = None
        self._SCPDURL = None

        self._serviceType = None
        self._serviceId = None

        self._validate_parameter_values = True

        self._default_variables = {}
        self._create_default_variables_from_list()

        # Initialize the evented variable sink
        super().__init__(self.SERVICE_EVENT_VARIABLES, state_lock=self._service_lock, sink_prefix=self.SERVICE_TYPE, auto_subscribe=True)

        self._logged_events = None
        if self.SERVICE_TYPE is not None:
            ctx = Context()
            logged_events_by_service = ctx.lookup(ContextPaths.UPNP_LOGGED_EVENTS)
            if logged_events_by_service is not None and self.SERVICE_TYPE in logged_events_by_service:
                self._logged_events = logged_events_by_service[self.SERVICE_TYPE]
        return

    @property
    def baseUrl(self) -> str:
        """
            Returns the base URL for working with a service.
        """
        return self._baseURL

    @property
    def controlURL(self) -> str:
        """
            Returns the control URL that is used making calls on actions on a device service.
        """
        return self._controlURL

    @property
    def device(self) -> "UpnpDevice":
        """
            Returns a reference to the parent device of this service.
        """
        return self._device_ref()

    @property
    def eventSubURL(self) -> str:
        """
            Returns the URL that is used to subscribe to events.
        """
        return self._eventSubURL

    @property
    def host(self) -> str:
        """
            The host associated with the parent device.
        """
        return self._host

    @property
    def SCPDURL(self) -> str:
        """
            The URL called to retrieve the service description document.
        """
        return self._SCPDURL

    @property
    def serviceId(self) -> str:
        """
            Returns the service ID
        """
        return self._serviceId

    @property
    def serviceType(self) -> str:
        """
            Returns the service Type ID.
        """
        return self._serviceType

    def call_action(self, action_name: str, arguments: dict = {}, auth: dict = None, headers: dict = {}, aspects: AspectsUPnP=DEFAULT_UPNP_CALL_ASPECTS):
        """
            Method utilize to make direct calls on a service for action APIs that are not published in a service description.

            :param action_name: The action name to make a call on.
            :param arguments: The arguments to pass to the action.
            :param auth: The authentication parameter to use when making a request on the remote action.
            :param headers: The headers to use when making the request on the remote action.

            :returns: Returns a dictionary which contains the returned response data.
        """
        rtnval = None

        completion_toctx = TimeoutContext(aspects.completion_timeout, aspects.completion_interval)
        monitor_delay = timedelta(seconds=aspects.monitor_delay)

        retry_logging_interval = aspects.retry_logging_interval
        allowed_error_codes = aspects.allowed_error_codes
        must_connect = aspects.must_connect

        this_thr = threading.current_thread()
        monmsg= "Thread failed to exit monitored scope. thid={} thname={} action_name={}".format(this_thr.ident, this_thr.name, action_name)

        if aspects.action_pattern == ActionPattern.SINGLE_CALL:
            with MonitoredScope("CALL_ACTION-SINGLE_CALL", monmsg, completion_toctx, notify_delay=monitor_delay) as _:
                rtnval = self._proxy_call_action(action_name, arguments=arguments, auth=auth, headers=headers, aspects=aspects)
        
        elif aspects.action_pattern == ActionPattern.SINGLE_CONNECTED_CALL:
            completion_toctx.mark_begin()
            retry_counter = 0

            while True:

                with MonitoredScope("CALL_ACTION-SINGLE_CONNECTED_CALL", monmsg, completion_toctx, notify_delay=monitor_delay) as _:
                    try:
                        rtnval = self._proxy_call_action(action_name, arguments=arguments, auth=auth, headers=headers, aspects=aspects)
                        break
                    except UpnpError as upnp_err:
                        # If we got a UpnpError, then the call to the device succeeded from a connectivity perspective.
                        # We log the error and raise an exception.
                        err_code = upnp_err.errorCode
                        err_description = upnp_err.errorDescription
                        err_extra = upnp_err.extra
                        if retry_counter % retry_logging_interval == 0:
                            dbg_msg_lines = [
                                "UpnpError: calling '{}' args={} attempt={} errCode={} errDescription={}".format(
                                    action_name, arguments, retry_counter + 1, err_code, err_description),
                                "EXTRA: {}".format(err_extra)
                            ]
                            dbg_msg = os.linesep.join(dbg_msg_lines)
                            logger.debug(dbg_msg)
                        raise
                    except Exception as xcpt:
                        err_msg_lines = [
                            "Exception raised by _proxy_call_action."
                        ]
                        err_msg_lines.extend(format_exc_lines())
                        err_msg = os.linesep.join(err_msg_lines)
                        logger.error(err_msg)

                if completion_toctx.final_attempt:
                    what_for="a single UPNP action call to transact"
                    details = [
                        "ACTION: %s" % action_name
                    ]
                    toerr = completion_toctx.create_timeout(what_for=what_for, detail=details)
                    raise toerr

                elif not completion_toctx.should_continue():
                    completion_toctx.mark_final_attempt()

                time.sleep(completion_toctx.interval)

                retry_counter += 1

        elif aspects.action_pattern == ActionPattern.DO_UNTIL_SUCCESS:
            completion_toctx.mark_begin()
            retry_counter = 0

            while True:

                with MonitoredScope("CALL_ACTION-DO_UNTIL_SUCCESS", monmsg, completion_toctx, notify_delay=monitor_delay) as _:
                    try:
                        rtnval = self._proxy_call_action(action_name, arguments=arguments, auth=auth, headers=headers, aspects=aspects)
                        break
                    except UpnpError as upnp_err:
                        # If we got a UpnpError, then the call to the device succeeded from a connectivity perspective.
                        # If there is an entry for the error in the allowed_upnp_errors, then we just log the error
                        # and allow a retry.
                        err_code = upnp_err.errorCode
                        err_description = upnp_err.errorDescription
                        err_extra = upnp_err.extra
                        if retry_counter % retry_logging_interval == 0:
                            dbg_msg_lines = [
                                "UpnpError: calling '{}' args={} attempt={} errCode={} errDescription={}".format(
                                    action_name, arguments, retry_counter + 1, err_code, err_description),
                                "EXTRA: {}".format(err_extra)
                            ]
                            dbg_msg = os.linesep.join(dbg_msg_lines)
                            logger.debug(dbg_msg)

                        # If we were given a list of allowed error codes and the error code is
                        # not in the list of allowed error codes, re-raise the UpnpError
                        if allowed_error_codes is not None and len(allowed_error_codes) > 0:
                            if err_code not in allowed_error_codes:
                                raise

                    except Exception as xcpt:
                        # Un-Successful calls, always allow a retry for the ActionPattern.DO_UNTIL_SUCCESS pattern.
                        err_msg_lines = [
                            "Exception raised by _proxy_call_action."
                        ]
                        err_msg_lines.extend(format_exc_lines())
                        err_msg = os.linesep.join(err_msg_lines)
                        logger.error(err_msg)

                        # If the must, connect flag was passed with ActionPattern.DO_UNTIL_SUCCESS then
                        # re-raise the exception after logging it.
                        if must_connect:
                            raise

                if completion_toctx.final_attempt:
                    what_for="UPNP action call success"
                    details = [
                        "ACTION: %s" % action_name
                    ]
                    toerr = completion_toctx.create_timeout(what_for=what_for, detail=details)
                    raise toerr

                elif not completion_toctx.should_continue():
                    completion_toctx.mark_final_attempt()

                time.sleep(completion_toctx.interval)

                retry_counter += 1

        elif aspects.action_pattern == ActionPattern.DO_WHILE_SUCCESS:
            completion_toctx.mark_begin()

            while True:

                with MonitoredScope("CALL_ACTION-DO_WHILE_SUCCESS", monmsg, completion_toctx, notify_delay=monitor_delay) as _:
                    try:
                        rtnval = self._proxy_call_action(action_name, arguments=arguments, auth=auth, headers=headers, aspects=aspects)
                    except UpnpError as upnp_err:
                        # If we got a UpnpError, then the call to the device succeeded from a connectivity perspective.
                        # If there is an entry for the error in the allowed_upnp_errors, then we just log the error
                        # and allow a retry.
                        err_code = upnp_err.errorCode
                        err_description = upnp_err.errorDescription
                        err_extra = upnp_err.extra
                        if retry_counter % retry_logging_interval == 0:
                            dbg_msg_lines = [
                                "UpnpError: calling '{}' args={} attempt={} errCode={} errDescription={}".format(
                                    action_name, arguments, retry_counter + 1, err_code, err_description),
                                "EXTRA: {}".format(err_extra)
                            ]
                            dbg_msg = os.linesep.join(dbg_msg_lines)
                            logger.debug(dbg_msg)
                        break
                    except Exception as xcpt:
                        if not must_connect:
                            break

                if completion_toctx.final_attempt:
                    what_for="UPNP action call failure"
                    details = [
                        "ACTION: %s" % action_name
                    ]
                    toerr = completion_toctx.create_timeout(what_for=what_for, detail=details)
                    raise toerr

                elif not completion_toctx.should_continue():
                    completion_toctx.mark_final_attempt()

                time.sleep(completion_toctx.interval)

        elif aspects.action_pattern == ActionPattern.DO_UNTIL_CONNECTION_FAILURE:
            completion_toctx.mark_begin()

            while True:

                with MonitoredScope("CALL_ACTION-DO_UNTIL_CONNECTION_FAILURE", monmsg, completion_toctx, notify_delay=monitor_delay) as _:
                    try:
                        rtnval = self._proxy_call_action(action_name, arguments=arguments, auth=auth, headers=headers, aspects=aspects)
                    except UpnpError as upnp_err:
                        # We only log UPNP errors
                        err_description = upnp_err.errorDescription
                        err_extra = upnp_err.extra
                        if retry_counter % retry_logging_interval == 0:
                            dbg_msg_lines = [
                                "UpnpError: calling '{}' args={} attempt={} errCode={} errDescription={}".format(
                                    action_name, arguments, retry_counter + 1, err_code, err_description),
                                "EXTRA: {}".format(err_extra)
                            ]
                            dbg_msg = os.linesep.join(dbg_msg_lines)
                            logger.debug(dbg_msg)
                    except Exception as xcpt:
                        break

                if completion_toctx.final_attempt:
                    what_for="UPNP connection failure"
                    details = [
                        "ACTION: %s" % action_name
                    ]
                    toerr = completion_toctx.create_timeout(what_for=what_for, detail=details)
                    raise toerr

                elif not completion_toctx.should_continue():
                    completion_toctx.mark_final_attempt()

                time.sleep(completion_toctx.interval)

        else:
            errmsg = "UpnpServiceProxy: Unknown ActionPattern encountered. action_pattern={}".format(aspects.action_pattern)
            raise AKitRuntimeError(errmsg) from None

        return rtnval

    def invalidate_subscription(self):
        """
            Called in order to invalidate the subscription(s) specified by scope.

            :param scope: The scope of the subscriptions to renew.  If not specified then all
                          subscriptions should be renewed.
        """
        
        for _ in self.yield_state_lock():
            self._subscription_id = None
            self._subscription_expiration = None

            for varkey in self._evented_variables:
                varobj = self._evented_variables[varkey]
                varobj.invalidate_subscription()

        return

    def lookup_default_variable(self, varname: str) -> Union[UpnpDefaultVar, None]:
        """
            Looks up the specified default variable.

            :param varname: The event name to find the :class:`UpnpDefaultVar` for.
        """
        varobj = None

        varkey = "{}/{}".format(self.SERVICE_TYPE, varname)

        for _ in self.yield_state_lock():
            if varkey in self._default_variables:
                varobj = self._default_variables[varkey]

        return varobj

    def renew_subscription(self):
        """
            Called in order to renew the subscription to the 
        """
        self.device.unsubscribe_to_events(self)
        sub_id, sub_expires = self.device.subscribe_to_events(self, renew=True)

        for _ in self.yield_state_lock():
            self._subscription_id = sub_id
            self._subscription_expiration = sub_expires
        
        return

    def trigger_auto_subscribe_from_variable(self, varkey: str):
        """
            Called in order to renew the subscription to the 

            :param varkey: The key for the variable that is triggering the auto-subscription process.
        """
        logger.debug("UpnpServiceProxy subscription renewal for {} triggered from variable {}".format(
            self.SERVICE_ID, varkey
        ))
        self.renew_subscription()
        return

    def _clear_subscription(self):

        for _ in self.yield_state_lock():
            self._subscription_id = None
            self._subscription_expiration = None
        
        return

    def _create_default_variable(self, event_name: str, data_type: Optional[str] = None, default: Optional[str] = None, allowed_list: Optional[list] = None, evented: bool=True):
        """
            Creates a default variable and stores a reference to it in the variables table.

            :param event_name: The name of the event variable to create.
            :param data_type: The type of the event variable to create.
            :param default: The default value to set the new event variable to.
        """

        variable_key = "{}/{}".format(self.SERVICE_TYPE, event_name)

        service_ref = weakref.ref(self)
        event_var = EventedVariable(variable_key, event_name, service_ref, data_type=data_type, default=default, allowed_list=allowed_list, evented=evented)
        self._default_variables[variable_key] = event_var

        return

    def _create_default_variables_from_list(self):
        """
            Called by the constructor to create the defalut variables listed in the SERVICE_DEFAULT_VARIABLES list on creation
            of the service proxy instance.  We pre-create the event variables because they can have default values and 
            we want to maintain consistent reference for the variables for the the life of the service instance.
        """
        for event_name in self.SERVICE_DEFAULT_VARIABLES:
            event_info = self.SERVICE_DEFAULT_VARIABLES[event_name]
            self._create_default_variable(event_name, **event_info)
        return

    def _proxy_link_service_to_device(self, device_ref: weakref.ReferenceType, service_description: UpnpDevice1Service):
        """
            Called to link a :class:`UpnpServiceProxy` instance to a :class:`UpnpDevice` instance.  The link to the parent
            device allows device users to find the service instance and to link the service proxy with the host it interacts
            with.  It is also utilized by the service proxy to setup event subscription callback routing in order to be
            able to route updates to the :class:`EventedVariable` variables managed by this service proxy.

            :param device_ref: A weak reference to the parent device that owns this device.
            :param service_description: The service description for this service.
        """

        device = device_ref()

        self._device_ref = device_ref

        self._host = device.host
        self._baseURL = device.URLBase

        self._controlURL = service_description.controlURL
        self._eventSubURL = service_description.eventSubURL
        self._SCPDURL = service_description.SCPDURL
        self._serviceId = service_description.serviceId
        self._serviceType = service_description.serviceType

        return

    def _proxy_set_call_parameters(self, host: str, baseURL: str, controlURL: str, eventSubURL: str, serviceId: Optional[str] = None, serviceType: Optional[str] = None):
        """
            Sets the call parameters that are used by the service for making calls on a remote service.

            :param host: The host of the remote service.
            :param baseURL: The base URL of the host and remote service.
            :param controlURL: The URL use to make calls on service actions.
            :param eventSubURL: The URL to use for creating service event variable subscriptions.
            :param serviceId: The service ID of the service.
            :param serviceType: The service Type ID of the service.
        """
        self._host = host

        if serviceId is None:
            serviceId = self.SERVICE_ID

        if serviceType is None:
            serviceType = self.SERVICE_TYPE

        self._baseURL = baseURL

        self._controlURL = controlURL
        self._eventSubURL = eventSubURL
        self._serviceId = serviceId
        self._serviceType = serviceType
        return

    def _proxy_call_action(self, action_name: str, arguments: dict = {}, auth: Optional[dict] = None, headers: dict = {}, aspects: AspectsUPnP=DEFAULT_UPNP_ASPECTS) -> dict:
        """
            Helper method utilize by derived service proxies to make calls on remote service actions.

            :param action_name: The action name to make a call on.
            :param arguments: The arguments to pass to the action.
            :param auth: The authentication parameter to use when making a request on the remote action.
            :param headers: The headers to use when making the request on the remote action.

            :returns: Returns a dictionary which contains the returned response data.
        """
        # pylint: disable=dangerous-default-value

        call_url = self.controlURL
        if self._baseURL is not None:
            call_url = self._baseURL + call_url

        call_body = self._soap_processor.create_request(action_name, arguments, typed=self.serviceType)

        call_headers = {
            'SOAPAction': '"%s#%s"' % (self.serviceType, action_name),
            'Host': self._host,
            'Content-Type': 'text/xml'
        }
        call_headers.update(headers)

        resp = None
        try:
            resp = requests.post(
                call_url,
                call_body,
                headers=call_headers,
                timeout=SOAP_TIMEOUT,
                auth=auth
            )
            resp.raise_for_status()
        except requests.exceptions.HTTPError:
            if resp is not None:
                # If the body of the error response contains XML then it should be a UPnP error,
                # extract the UPnP error information and raise a UpnpError
                content_type = resp.headers["CONTENT-TYPE"]
                if content_type.find('text/xml') == -1:
                    raise
        except:
            errmsg = traceback.format_exc()
            print(errmsg)
            raise

        resp_content = resp.content.strip()

        resp_dict = None

        status_code = resp.status_code
        if status_code >= 200 and status_code < 300: # pylint: disable=chained-comparison
            resp_dict = self._soap_processor.parse_response(action_name, resp_content, typed=self.serviceType)
        else:
            errorCode, errorDescription = self._soap_processor.parse_response_error_for_upnp(action_name, resp_content, status_code, typed=self.serviceType)
            raise UpnpError(errorCode, errorDescription, "host=%s action=%s args=%s" % (self._host, action_name, repr(arguments)))

        return resp_dict

    def _update_event_variables(self, sender_ip, usn_dev, propertyNodeList):
        """
            Helper method called during the processing of a subscription callback in order
            to update all of the event variables for this service instance.

            :param propertyNodeList: An XML :class:`Element` object that contains a list
                                     of child elements for each event variable.
        """

        for _ in self.yield_state_lock():
            for propNodeOuter in propertyNodeList:
                # Get the first node of the outer property node
                propNode = propNodeOuter.getchildren()[0]

                event_name = propNode.tag
                event_value = propNode.text
                if event_value is None:
                    event_value = ""

                if self._logged_events is not None and event_name in self._logged_events:
                        infomsg = "UPNP event update for {}/{}/{} from {}{}    VALUE: {}".format(
                            usn_dev, self.SERVICE_TYPE, event_name, sender_ip, os.linesep, event_value)
                        logger.debug(infomsg)

                try:
                    self.update_event_variable(event_name, event_value, sink_locked=True)
                except KeyError:
                    logger.debug("UpnpServiceProxy: Received value for unknown event host=%s event=%s value=%r" % (sender_ip, event_name, event_value))

        return



"""
.. module:: dnsrecord
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Contains the DnsRecord object which serves as the base object type for other DNS record types.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""


__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Optional, Union, cast, TYPE_CHECKING

import socket
import time

from akit.exceptions import AKitAbstractMethodError

from akit.networking.interfaces import is_ipv6_address
from akit.interop.dns.dnsconst import DnsLiftimePercent

from akit.xtime import current_time_millis

from akit.interop.dns.dnsconst import DnsRecordClass
from akit.interop.dns.dnsconst import DnsRecordType

if TYPE_CHECKING:
    from akit.interop.dns.dnsincoming import DnsIncoming
    from akit.interop.dns.dnspacketwriter import DnsPacketWriter

class DnsRecord:
    """
        A DNS record is a DNS entry that has a Time To Live or TTL attached.

                                         1  1  1  1  1  1
           0  1  2  3  4  5  6  7  8  9  0  1  2  3  4  5
        +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        |                                               |
        /                                               /
        /                      NAME                     /
        |                                               |
        +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        |                      TYPE                     |
        +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        |                     CLASS                     |
        +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        |                      TTL                      |
        |                                               |
        +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
        |                   RDLENGTH                    |
        +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--|
        /                     RDATA                     /
        /                                               /
        +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+

    """
    def __init__(self, name: str, rtype: DnsRecordType, rclass: DnsRecordClass, ttl: Union[float, int]) -> None:
        self._key = name.lower()
        self._name = name
        self._rtype = rtype
        self._rclass = (rclass & DnsRecordClass.MASK)
        self._ttl = ttl

        self._unique = (rclass & DnsRecordClass.UNIQUE) != 0
        self._updated = current_time_millis()

        self._when_expires = self._compute_time_marker(DnsLiftimePercent.Expired)
        self._when_refresh = self._compute_time_marker(DnsLiftimePercent.Refresh)
        self._when_stale = self._compute_time_marker(DnsLiftimePercent.Stale)
        return

    @property
    def key(self):
        return self._key

    @property
    def name(self):
        return self._name

    @property
    def rclass(self):
        return self._rclass

    @property
    def rtype(self):
        return self._rtype

    @property
    def ttl(self):
        return self._ttl

    @property
    def unique(self):
        return self._unique

    @property
    def updated(self):
        return self._updated

    def as_dns_string(self, other: Optional[str] = None) -> str:
        """
            Creates a DNS string representation of the DNS record.
        """
        remaining_ttl = int(self.get_remaining_ttl(current_time_millis()))
        other = "%s/%s,%s" % (self.ttl, remaining_ttl, other)
        
        dnsstr = "%s[%s,%s" % ('record', self._rtype.name, self._rclass.name)

        if self._unique:
            dnsstr += "-unique,"
        else:
            dnsstr += ","

        dnsstr += self._name

        if other is not None:
            dnsstr += "]=%s" % other
        else:
            dnsstr += "]"

        return dnsstr

    def get_remaining_ttl(self, now: float) -> float:
        """
            Gets the remaining time to live (TTL) in seconds.
        """
        rval = max(0, (self._when_expires - now) / 1000)
        return rval

    def is_expired(self, now: Optional[float]) -> bool:
        """
            Returns true if this record is expired.
        """
        if now is None:
            now = time.time()

        bval = self._when_expires <= now
        return bval

    def is_stale(self, now: Optional[float]) -> bool:
        """
            Returns true if this record is stale.
        """
        if now is None:
            now = time.time()

        bval = self._when_stale <= now
        return bval

    def suppressed_by(self, msg: 'DnsIncoming') -> bool:
        """
            Returns true if any answer in a message can suffice for the information held in this
            record.
        """
        rtn_val = False
        for record in msg.answers:
            if self.suppressed_by_answer(record):
                rtn_val = True
                break
        return rtn_val

    def suppressed_by_answer(self, other: 'DnsRecord') -> bool:
        """
            Returns true if another record has same name, type and class, and if its TTL is at least
            half of this record's.
        """
        rtnval = self == other and other.ttl > (self.ttl / 2)
        return rtnval

    def update_ttl(self, other: 'DnsRecord') -> None:
        """
            Updates the time to live TTL from the specified record.
        """
        self._updated = other._updated
        self._ttl = other._ttl
        self._when_expires = self._compute_time_marker(DnsLiftimePercent.Expired)
        self._when_refresh = self._compute_time_marker(DnsLiftimePercent.Refresh)
        self._when_stale = self._compute_time_marker(DnsLiftimePercent.Stale)
        return

    def write(self, out: 'DnsPacketWriter') -> None:
        """
            Abstract method
        """
        raise AKitAbstractMethodError from None

    def _compute_time_marker(self, percent: int) -> float:
        marker = self._updated + (percent * self._ttl * 10)
        return marker

    def __eq__(self, other: 'DnsRecord') -> bool:
        """
            Equality test on name, record type, and record class
        """
        iseq = False
        if isinstance(other, type(self)):
            if self._rtype == other._rtype and self._rclass and other._rclass:
                iseq = self._name == other._name
        return iseq

    def __ne__(self, other: 'DnsRecord') -> bool:
        """
            Non-equality test
        """
        isne = not self.__eq__(other)
        return isne

    def __str__(self) -> str:
        """
            String representation
        """
        dnsstr = self.as_dns_string()
        return dnsstr



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class RATAConfig1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:RATAConfig:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'RATAConfig1'

    SERVICE_DEFAULT_VARIABLES = {
        "CredentialDelivery": { "data_type": "string", "default": None, "allowed_list": None},
        "ProfileList": { "data_type": "string", "default": None, "allowed_list": None},
        "SystemInfo": { "data_type": "string", "default": None, "allowed_list": None},
        "TransportAgentCapabilities": { "data_type": "string", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {
        "CredentialsList": { "data_type": "string", "default": None, "allowed_list": None},
    }

    def action_AddProfile(self, NewProfileConfigInfo, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the AddProfile action.
        """
        arguments = {
            "NewProfileConfigInfo": NewProfileConfigInfo,
        }

        self.call_action("AddProfile", arguments=arguments, aspects=aspects)

        return

    def action_DeleteProfile(self, ProfileID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the DeleteProfile action.
        """
        arguments = {
            "ProfileID": ProfileID,
        }

        self.call_action("DeleteProfile", arguments=arguments, aspects=aspects)

        return

    def action_EditProfile(self, ProfileID, UpdatedProfileConfigInfo, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the EditProfile action.
        """
        arguments = {
            "ProfileID": ProfileID,
            "UpdatedProfileConfigInfo": UpdatedProfileConfigInfo,
        }

        self.call_action("EditProfile", arguments=arguments, aspects=aspects)

        return

    def action_GetCredentialsList(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetCredentialsList action.

            :returns: "CurrentCredentialsList"
        """
        arguments = { }

        out_params = self.call_action("GetCredentialsList", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentCredentialsList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetProfileConfigInfo(self, ProfileID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetProfileConfigInfo action.

            :returns: "ProfileConfigInfo"
        """
        arguments = {
            "ProfileID": ProfileID,
        }

        out_params = self.call_action("GetProfileConfigInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("ProfileConfigInfo",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetProfileList(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetProfileList action.

            :returns: "ProfileList"
        """
        arguments = { }

        out_params = self.call_action("GetProfileList", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("ProfileList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetSupportedCredentialDelivery(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetSupportedCredentialDelivery action.

            :returns: "SupportedCredentialDelivery"
        """
        arguments = { }

        out_params = self.call_action("GetSupportedCredentialDelivery", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("SupportedCredentialDelivery",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetTransportAgentCapabilities(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetTransportAgentCapabilities action.

            :returns: "TransportAgentCapabilities"
        """
        arguments = { }

        out_params = self.call_action("GetTransportAgentCapabilities", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("TransportAgentCapabilities",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class WANCableLinkConfig1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:WANCableLinkConfig:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'WANCableLinkConfig1'

    SERVICE_DEFAULT_VARIABLES = {
        "BPIEncryptionEnabled": { "data_type": "boolean", "default": None, "allowed_list": None},
        "CableLinkConfigState": { "data_type": "string", "default": None, "allowed_list": "['notReady', 'dsSyncComplete', 'usParamAcquired', 'rangingComplete', 'ipComplete', 'todEstablished', 'paramTransferComplete', 'registrationComplete', 'operational', 'accessDenied']"},
        "ConfigFile": { "data_type": "string", "default": None, "allowed_list": None},
        "DownstreamFrequency": { "data_type": "ui4", "default": None, "allowed_list": None},
        "DownstreamModulation": { "data_type": "string", "default": None, "allowed_list": "['64QAM', '256QAM']"},
        "LinkType": { "data_type": "string", "default": None, "allowed_list": "['Ethernet']"},
        "TFTPServer": { "data_type": "string", "default": None, "allowed_list": None},
        "UpstreamChannelID": { "data_type": "ui4", "default": None, "allowed_list": None},
        "UpstreamFrequency": { "data_type": "ui4", "default": None, "allowed_list": None},
        "UpstreamModulation": { "data_type": "string", "default": None, "allowed_list": "['QPSK', '16QAM']"},
        "UpstreamPowerLevel": { "data_type": "ui4", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {}

    def action_GetBPIEncryptionEnabled(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetBPIEncryptionEnabled action.

            :returns: "NewBPIEncryptionEnabled"
        """
        arguments = { }

        out_params = self.call_action("GetBPIEncryptionEnabled", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewBPIEncryptionEnabled",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetCableLinkConfigInfo(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetCableLinkConfigInfo action.

            :returns: "NewCableLinkConfigState", "NewLinkType"
        """
        arguments = { }

        out_params = self.call_action("GetCableLinkConfigInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewCableLinkConfigState", "NewLinkType",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetConfigFile(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetConfigFile action.

            :returns: "NewConfigFile"
        """
        arguments = { }

        out_params = self.call_action("GetConfigFile", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewConfigFile",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetDownstreamFrequency(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetDownstreamFrequency action.

            :returns: "NewDownstreamFrequency"
        """
        arguments = { }

        out_params = self.call_action("GetDownstreamFrequency", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewDownstreamFrequency",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetDownstreamModulation(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetDownstreamModulation action.

            :returns: "NewDownstreamModulation"
        """
        arguments = { }

        out_params = self.call_action("GetDownstreamModulation", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewDownstreamModulation",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetTFTPServer(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetTFTPServer action.

            :returns: "NewTFTPServer"
        """
        arguments = { }

        out_params = self.call_action("GetTFTPServer", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewTFTPServer",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetUpstreamChannelID(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetUpstreamChannelID action.

            :returns: "NewUpstreamChannelID"
        """
        arguments = { }

        out_params = self.call_action("GetUpstreamChannelID", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewUpstreamChannelID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetUpstreamFrequency(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetUpstreamFrequency action.

            :returns: "NewUpstreamFrequency"
        """
        arguments = { }

        out_params = self.call_action("GetUpstreamFrequency", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewUpstreamFrequency",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetUpstreamModulation(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetUpstreamModulation action.

            :returns: "NewUpstreamModulation"
        """
        arguments = { }

        out_params = self.call_action("GetUpstreamModulation", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewUpstreamModulation",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetUpstreamPowerLevel(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetUpstreamPowerLevel action.

            :returns: "NewUpstreamPowerLevel"
        """
        arguments = { }

        out_params = self.call_action("GetUpstreamPowerLevel", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewUpstreamPowerLevel",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args



"""
.. module:: dbio
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Contains helper functions for working with specific databases

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import contextlib
import traceback

from akit.exceptions import AKitConfigurationError

from akit.datum.orm import AutomationBase

from akit.datum.dbconnection import DatabaseConnectionFactory, lookup_database_connection_factory

def database_exists(conn_factory: DatabaseConnectionFactory, dbname: str):
    """
        Checks to see if the specified database exists.
    """
    engine = conn_factory.create_engine(dbname="postgres", echo=True)
    result = engine.execute("SELECT 1 AS result FROM pg_database WHERE datname='%s'" % dbname)
    return result.rowcount > 0

def database_create(conn_factory: DatabaseConnectionFactory, dbname):
    """
        Creates the specified database.
    """
    engine = conn_factory.create_engine(dbname="postgres", echo=True)

    conn = engine.connect()
    try:
        conn.connection.set_isolation_level(0)
        result = conn.execute("CREATE DATABASE %s" % dbname)
    except Exception as xcpt:
        print(traceback.format_exc())
        raise
    finally:
        conn.close()

    return

def create_automation_database(db_profile_name: str):
    """
        Creates the 'automation' database.
    """
    dbname = "automation"
    conn_factory = lookup_database_connection_factory(db_profile_name)

    if not database_exists(conn_factory, dbname):
        database_create(conn_factory, dbname)

    engine = conn_factory.create_engine(dbname=dbname, echo=True)

    # TODO: Solidify the data model and the metadata base used
    # to create the database
    AutomationBase.metadata.create_all(engine, checkfirst=True)

    return

def open_automation_database(db_profile_name: str):
    """
        Opens the 'automation' postgresql database.
    """
    dbname = 'automation'

    conn_factory = lookup_database_connection_factory(db_profile_name)
    if conn_factory is None:
        errmsg = "'open_automation_database' could not get a connection factory for profile={}".format(
            db_profile_name
        )
        raise AKitConfigurationError(errmsg)

    engine = conn_factory.create_engine(dbname=dbname, echo=True)

    return engine

def reset_automation_database(db_profile_name: str):
    """
        Reset the 'automation' database.
    """
    dbname = "automation"
    conn_factory = lookup_database_connection_factory(db_profile_name)

    if database_exists(conn_factory, dbname):
    
        engine = conn_factory.create_engine(dbname=dbname, echo=True)

        with contextlib.closing(engine.connect()) as con:
            trans = con.begin()
            con.execute('TRUNCATE {} RESTART IDENTITY;'.format(
                ','.join(table.name 
                        for table in reversed(AutomationBase.metadata.sorted_tables))))
            trans.commit()

    return




"""
.. module:: createenvironment
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: A module that provides the CreateEnvironment task class which implements
               the execution of environment creation commands.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Dict, Optional

import os
import shlex
import stat
import subprocess

from akit.xformatting import indent_lines

from akit.workflow.tasks.taskbase import TaskBase

class CreateEnvironment(TaskBase):
    """
        A task class that is used to create environment variables that persist and
        are consumable over the entire workflow.
    """

    def __init__(self, ordinal: int, label: str, task_info: dict, logger):
        super(CreateEnvironment, self).__init__(ordinal, label, task_info, logger)
        self._variables = task_info["variables"]
        return

    @property
    def variables(self) -> Dict[str, str]:
        return self._variables

    def execute(self, parameters: Optional[dict]=None, topology: Optional[dict]=None, **kwargs) -> int:

        self._logger.info("STEP: %s - %s" % (self._label, self._ordinal))

        status_code = 0

        for var_info in self._variables:
            var_name = var_info["name"]
            var_command = var_info["command"]

            stdout, stderr, exit_code = None, None, None

            # If "VIRTUAL_ENV" is set then we are running in a virtual environment, setup to run the shell
            # command in the same virtual environment we are currently running in.
            if "VIRTUAL_ENV" in os.environ:
                virtual_env = os.environ["VIRTUAL_ENV"]
                virtual_env_bin = os.path.join(virtual_env, 'bin')
                virtual_env_activate_script = os.path.join(virtual_env_bin, 'activate')
                virtual_env_cmd_script = os.path.join(virtual_env_bin, 'virtual_env_cmd')
                if not os.path.exists(virtual_env_cmd_script):
                    with open(virtual_env_cmd_script, 'w') as vecsf:
                        vecsf.write("#!/usr/bin/env bash\n")
                        vecsf.write("cd {}\n".format(virtual_env_bin))
                        vecsf.write("source {}\n".format(virtual_env_activate_script))
                        vecsf.write("\"$@\"\n")
                    os.chmod(virtual_env_cmd_script, stat.S_IREAD | stat.S_IWRITE | stat.S_IEXEC)

                proc = subprocess.Popen([virtual_env_cmd_script, var_command], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                stdout, stderr = proc.communicate()
                exit_code = proc.wait()

            else:
                proc = subprocess.Popen([var_command], stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
                stdout, stderr = proc.communicate()
                exit_code = proc.wait()

            if exit_code == 0:
                var_val = stdout.strip()
                os.environ[var_name] = var_val
            else:
                errmsg_lines = [
                    "Error creating environment variable '{}' with:".format(var_name),
                    "COMMAND: {}".format(var_command),
                    "STDERR: {}".format(stderr)
                ]
                errmsg = os.linesep.join(errmsg_lines)
                self._logger.error(errmsg)

        return status_code




from typing import Any, Callable, Dict, Optional, Type, Union

import inspect

from akit.testing.testplus.registration.sourcebase import SourceBase
from akit.testing.testplus.resourcelifespan import ResourceLifespan

class ParameterOrigin:

    def __init__(self, originating_scope: str, identifier: str, source: SourceBase, life_span: ResourceLifespan, constraints: Optional[Dict]):
        self._originating_scope = originating_scope
        self._identifier = identifier
        self._source = source
        self._life_span = life_span
        self._constraints = constraints
        return

    @property
    def constraints(self) -> Union[dict, None]:
        """
            Returns the most applicable constraints assoceated with this resource subscription.  If the
            subscription constraints are set then they will be used.  If the subscription constraints are
            not set and the source constraints are set, then the source constraints will be returned. If
            no constraints are applied to the subscription or the source, 'None' will be returned.
        """
        cval = None
        if self._constraints is not None:
            cval = self._constraints
        elif self._source.constraints is not None:
            cval = self._source.constraints
        return cval

    @property
    def identifier(self) -> str:
        return self._identifier

    @property
    def life_span(self) -> ResourceLifespan:
        return self._life_span

    @property
    def originating_scope(self) -> str:
        return self._originating_scope

    @property
    def source_function(self) -> Callable:
        return self._source.source_function

    @property
    def source_function_name(self) -> str:
        return self._source.source_function.__name__

    @property
    def source_signature(self) -> inspect.Signature:
        return self._source.source_signature

    @property
    def source_id(self) -> str:
        idstr = self._source.source_id
        return idstr

    @property
    def source_module_name(self) -> str:
        return self._source.module_name

    @property
    def source_resource_type(self) -> Type:
        return self._source.resource_type

    def describe_source(self):
        descstr = self.source_id
        cval = self.constraints
        if cval is not None:
            descstr += " constraints={}".format(repr(cval))
        return descstr

    def generate_call(self):
        call_arg_str = ""
        call_args = [param for param in self.source_signature.parameters]
        if len(call_args) > 0:
            call_arg_str = ", ".join(call_args)
        call_str = "{}({})".format(self._source.source_function.__name__, call_arg_str)
        return call_str


from typing import Callable, Type

from akit.testing.testplus.registration.sourcebase import SourceBase

class UnknownSource(SourceBase):
    """
        The :class:`UnknownSource` object is used when a parameter is declared for a
        test method of other resource function that cannot be resolved to a source
        function.  The :class:`UnknownSource` enables the test parameters to be queried
        by acting as a placeholder.
    """
    def __init__(self, source_func: Callable, resource_type: Type, constraints: dict):
        SourceBase.__init__(self, source_func, resource_type, constraints)
        return
